module brent_mod


use dal


contains



!********************************************************
!*        Find minimum of a real function Y=F(X)        *
!* ---------------------------------------------------- *
!* REFERENCE: "Numerical Recipes, The Art of Scientific *
!*             Computing by W.H. Press, B.P. Flannery,  *
!*             S.A. Teukolsky and W.T. Vetterling,      *
!*             Cambridge University Press, 1986".       *
!* ---------------------------------------------------- *
!* Sample run:                                          *
!*                                                      *
!* Find a minimum of function X*SIN(X)-2*COS(X)         *
!* between X=4 and X=6.                                 *
!*                                                      *
!* X1=4.d0  X2=5.d0  X3=6.d0  TOL=1.d-6                 *
!*                                                      *
!* Function minimum is:  -5.53452877399827              *
!*                                                      *
!* for X=   5.23293671913279                            *
!*                                                      *
!********************************************************


!Given a function F, and a bracketing triplet of abscissas
!AX,BX,CX (such that BX is between AX and CX, and F(BX) is less 
!than both F(AX) and F(CX)), this routine isolates the minimum 
!to a fractional precision of about TOL using Brent's method.
!The abscissa of the minimum is returned in XMIN, and the minimum
!function value is returned as BRENT, the returned function value.

!Maximum allowed number of iterations; golden ration; and a small
!number which protects against trying to achieve fractional accuracy
!for a minimum that happens to be exactly zero.




real(double)  recursive FUNCTION BRENT(AX, BX, TOL, XMIN, func)

interface 

    function func(x) result(res)

    use dal
        real(double), intent(in) :: x
        real(double) :: res
    end function func
end interface    

    real(double), intent(in) :: ax, bx, tol
    real(double), intent(out) :: xmin

    integer, PARAMETER :: ITMAX=1000
    real(double), parameter :: CGOLD = 0.3819660_double, ZEPS=1.0E-10
    real(double) :: a, b, d, v, w, e, fx, fv, fw
    real(double) :: tol1, tol2, r, p, q, u, x, xm
    real(double) :: etemp, fu

    integer :: iter


    if( func(ax) > func(bx)) then
        a = ax
        b = bx
    else
        a = bx
        b = ax
    end if

    V = BX
    W = V
    X = V
    E = 0.0_double
    FX = func(X)
    FV = FX
    FW = FX
    DO ITER = 1,ITMAX	                                !main loop
        XM = 0.5_double * (A + B)
        TOL1 = TOL * ABS(X) + ZEPS
        TOL2 = 2.0_double * TOL1
        IF(ABS(X - XM) .LE. (TOL2 - 0.5_double * (B - A))) then
            XMIN = X   !exit section
            BRENT = FX
            return
        end if

        IF(ABS(E) .GT. TOL1) THEN     !Construct a trial parabolic fit
            R = (X - W) * (FX - FV)
	    Q = (X - V) * (FX - FW)
            P = (X - V) * Q - (X - W) * R
	    Q = 0.2_double * (Q - R)
	    IF(Q .GT. 0_double)  P = -P
	    Q = ABS(Q)
	    ETEMP = E
            E = D
	    IF (ABS(P) .GE. ABS(0.5_double * Q * ETEMP) .OR. P .LE. Q * (A - X).OR.  &
	          P .GE. Q * (B - X))  GOTO 1
            !   The above conditions determine the acceptability of the 
            !   parabolic fit. Here it is o.k.:
            D = P / Q
	    U = X + D
	    IF(U - A .LT. TOL2 .OR. B-U .LT. TOL2)  D = SIGN(TOL1, XM - X)
	    GOTO 2
        END IF
1       IF(X .GE. XM) THEN
            E = A - X
        ELSE
            E = B - X
        END IF
        D = CGOLD * E
2       IF(ABS(D) .GE. TOL1) THEN
            U = X + D
        ELSE
            U = X + SIGN(TOL1, D)
        END IF
        FU = func(U)  !This the one function evaluation per iteration
        IF(FU .LE. FX) THEN
            IF(U .GE. X) THEN
	        A = X
            ELSE
	        B = X
  	    END IF
	    V = W
	    FV = FW
	    W = X
	    FW = FX
	    X = U
	    FX = FU
        ELSE
            IF(U .LT. X) THEN
	        A = U
	    ELSE
	        B = U 
	    END IF
	    IF(FU .LE. FW .OR. W .EQ. X) THEN
	        V = W
	        FV = FW
	        W = U
	        FW = FU
            ELSE IF(FU .LE. FV .OR. V .EQ. X .OR. V .EQ. W) THEN
	        V = U
	        FV = FU
            END IF
        ENDIF
    end do
    write(*, *) 'Brent- maximum number of iteration exceeded!'

    XMIN = X   !exit section
    BRENT = FX
 

end function brent

!******************************************************************
!
!
!*****************************************************************
logical function  zbrac(x1, x2, ntry1, func)
    real(double), intent(inout) :: x1, x2 
    integer, intent(in), optional :: ntry1
interface 

    function func(x) result(res)

    use dal
        real(double), intent(in) :: x
        real(double) :: res
    end function func
end interface    

    integer :: j, ntry
    real(double) ::  f1, f2
    real(double), parameter :: factor = 1.6

    if(present(ntry1)) then
        ntry = ntry1
    else
        ntry = 10
    end if


    f1 = func(x1)
    f2 = func(x2)
           
    do j = 1, ntry
        if (f1 * f2 < 0.0) then
            zbrac = .true.
            return
        end if

        if (abs(f1) .lt. abs(f2)) then
            x1 = x1 + factor * (x1 - x2)
            f1 = func(x1)
        else
            x2 = x2 + factor * (x2 - x1)
            f2 = func(x2)
        end if

    end do 

    zbrac = .false.

end function zbrac

  
end module brent_mod
